
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "mpg123.h"

static GtkWidget *window = NULL, *vbox, *frame, *table;
static GtkWidget *title_label, *title_entry;
static GtkWidget *artist_label, *artist_entry;
static GtkWidget *album_label, *album_entry;
static GtkWidget *year_label, *year_entry;
static GtkWidget *comment_label, *comment_entry;
static GtkWidget *genre_label, *genre_combo;
static GtkWidget *bbox, *save, *remove_id3, *cancel;

static GList *genre_list = NULL;
static gchar *current_filename = NULL;

static gchar *id3_genres[] =
{
	"Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk", "Grunge", "Hip-Hop",
	"Jazz", "Metal", "New Age", "Oldies", "Other", "Pop", "R&B", "Rap", "Reggae", "Rock",
	"Techno", "Industrial", "Alternative", "Ska", "Death Metal", "Pranks", "Soundtrack",
	"Euro-Techno", "Ambient", "Trip-Hop", "Vocal", "Jazz+Funk", "Fusion", "Trance",
	"Classical", "Instrumental", "Acid", "House", "Game", "Sound Clip", "Gospel", "Noise",
   "Alt", "Bass", "Soul", "Punk", "Space", "Meditative", "Instrumental Pop",
   "Instrumental Rock", "Ethnic", "Gothic", "Darkwave", "Techno-Industrial",
	"Electronic", "Pop-Folk", "Eurodance", "Dream", "Southern Rock", "Comedy", "Cult",
	"Gangsta Rap", "Top 40", "Christian Rap", "Pop/Funk", "Jungle", "Native American",
	"Cabaret", "New Wave", "Psychedelic", "Rave", "Showtunes", "Trailer", "Lo-Fi",
	"Tribal", "Acid Punk", "Acid Jazz", "Polka", "Retro", "Musical", "Rock & Roll",
	"Hard Rock", "Folk", "Folk/Rock", "National Folk", "Swing", "Fast-Fusion", "Bebob",
     "Latin", "Revival", "Celtic", "Bluegrass", "Avantgarde", "Gothic Rock",
	"Progressive Rock", "Psychedelic Rock", "Symphonic Rock", "Slow Rock", "Big Band",
	"Chorus", "Easy Listening", "Acoustic", "Humour", "Speech", "Chanson", "Opera",
"Chamber Music", "Sonata", "Symphony", "Booty Bass", "Primus", "Porn Groove",
	"Satire", "Slow Jam", "Club", "Tango", "Samba", "Folklore", "Ballad", "Power Ballad",
"Rhythmic Soul", "Freestyle", "Duet", "Punk Rock", "Drum Solo", "A Cappella",
	"Euro-House", "Dance Hall", "Goa", "Drum & Bass", "Club-House", "Hardcore", "Terror",
	"Indie", "BritPop", "Negerpunk", "Polsk Punk", "Beat", "Christian Gangsta Rap",
	"Heavy Metal", "Black Metal", "Crossover", "Contemporary Christian",
	"Christian Rock", "Merengue", "Salsa", "Thrash Metal", "Anime", "JPop", "Synthpop"
};

struct id3tag
{
	char tag[3];		/* always "TAG": defines ID3v1 tag 128 bytes before EOF */
	char title[30];
	char artist[30];
	char album[30];
	char year[4];
	char comment[30];
	unsigned char genre;
};

static void show_dialog(gchar * title, gchar * text)
{
	GtkWidget *dialog, *label, *button;

	dialog = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(dialog), title);
	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
	gtk_container_border_width(GTK_CONTAINER(dialog), 5);
	label = gtk_label_new(text);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), label, TRUE, TRUE, 0);
	gtk_widget_show(label);

	button = gtk_button_new_with_label("Close");
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(button, 75, -1);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
	gtk_widget_show(button);
	gtk_widget_grab_default(button);

	gtk_widget_show(dialog);
}

static guint
       strip_spaces(char *src, size_t n)
/* strips trailing spaces from string of length n
   returns length of adjusted string */
{
	gchar *space = NULL,	// last space in src
	     *start = src;

	while (n--)
		switch (*src++)
		{
			case '\0':
				n = 0;	// breaks out of while loop

				src--;
				break;
			case ' ':
				if (space == NULL)
					space = src - 1;
				break;
			default:
				space = NULL;	// don't terminate intermediate spaces

				break;
		}
	if (space != NULL)
	{
		src = space;
		*src = '\0';
	}
	return src - start;
}

static void set_entry_tag(GtkEntry * entry, gchar * tag, gint length)
{
	gint stripped_len;
	gchar *text;

	stripped_len = strip_spaces(tag, length);
	text = g_strdup_printf("%-*.*s", stripped_len, stripped_len, tag);
	gtk_entry_set_text(entry, text);
	g_free(text);
}

static void get_entry_tag(GtkEntry * entry, gchar * tag, gint length)
{
	gchar *text;

	text = gtk_entry_get_text(entry);
	memset(tag, ' ', length);
	memcpy(tag, text, strlen(text) > length ? length : strlen(text));
}

static gint find_genre_id(gchar * text)
{
	gint i;

	for (i = 0; i < sizeof (id3_genres) / sizeof (*id3_genres); i++)
	{
		if (!strcmp(id3_genres[i], text))
			return i;
	}
	return 0;
}

static gint genre_comp_func(gconstpointer a, gconstpointer b)
{
	return strcasecmp(a, b);
}

static void save_cb(GtkWidget * w, gpointer data)
{
	gint fd;
	struct id3tag tag;

	if ((fd = open(current_filename, O_RDWR)) != -1)
	{
		lseek(fd, -128, SEEK_END);
		read(fd, &tag, sizeof (struct id3tag));

		if (!strncmp(tag.tag, "TAG", 3))
			lseek(fd, -128, SEEK_END);
		else
			lseek(fd, 0, SEEK_END);
		tag.tag[0] = 'T';
		tag.tag[1] = 'A';
		tag.tag[2] = 'G';
		get_entry_tag(GTK_ENTRY(title_entry), tag.title, 30);
		get_entry_tag(GTK_ENTRY(artist_entry), tag.artist, 30);
		get_entry_tag(GTK_ENTRY(album_entry), tag.album, 30);
		get_entry_tag(GTK_ENTRY(year_entry), tag.year, 4);
		get_entry_tag(GTK_ENTRY(comment_entry), tag.comment, 30);
		tag.genre = find_genre_id(gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(genre_combo)->entry)));
		if (write(fd, &tag, sizeof (struct id3tag)) != sizeof (struct id3tag))
			       show_dialog("File Info", "\n    Couldn't write tag!    \n");

		close(fd);
	}
	else
		show_dialog("File Info", "\n    Couldn't write tag!    \n");
	gtk_widget_destroy(window);
}

static void remove_id3_cb(GtkWidget * w, gpointer data)
{
	gint fd, len;
	struct id3tag tag;

	if ((fd = open(current_filename, O_RDWR)) != -1)
	{
		len = lseek(fd, -128, SEEK_END);
		read(fd, &tag, sizeof (struct id3tag));

		if (!strncmp(tag.tag, "TAG", 3))
		{
			if (ftruncate(fd, len))
				show_dialog("File Info", "\n    Couldn't remove tag!    \n");
		}
		else
			show_dialog("File Info", "\n    No tag to remove!    \n");
		close(fd);
	}
	else
		show_dialog("File Info", "\n    Couldn't remove tag!    \n");
	gtk_widget_destroy(window);
}

void mpg123_file_info_box(char *filename)
{
	gint i, len;
	struct id3tag tag;
	gint fd;
	gchar *tmp, *name, *title;

	if (!strncasecmp(filename, "http://", 7))
		return;

	if (!window)
	{
		window = gtk_window_new(GTK_WINDOW_DIALOG);
		gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &window);
		gtk_container_set_border_width(GTK_CONTAINER(window), 10);

		vbox = gtk_vbox_new(FALSE, 10);
		gtk_container_add(GTK_CONTAINER(window), vbox);

		frame = gtk_frame_new("ID3 Tag:");
		gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);

		table = gtk_table_new(4, 5, FALSE);
		gtk_container_set_border_width(GTK_CONTAINER(table), 5);
		gtk_container_add(GTK_CONTAINER(frame), table);

		title_label = gtk_label_new("Title:");
		gtk_misc_set_alignment(GTK_MISC(title_label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), title_label, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 5, 5);
		gtk_widget_show(title_label);

		title_entry = gtk_entry_new_with_max_length(30);
		gtk_table_attach(GTK_TABLE(table), title_entry, 1, 4, 0, 1, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);
		gtk_widget_show(title_entry);

		artist_label = gtk_label_new("Artist:");
		gtk_misc_set_alignment(GTK_MISC(artist_label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), artist_label, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 5, 5);
		gtk_widget_show(artist_label);

		artist_entry = gtk_entry_new_with_max_length(30);
		gtk_table_attach(GTK_TABLE(table), artist_entry, 1, 4, 1, 2, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);
		gtk_widget_show(artist_entry);

		album_label = gtk_label_new("Album:");
		gtk_misc_set_alignment(GTK_MISC(album_label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), album_label, 0, 1, 2, 3, GTK_FILL, GTK_FILL, 5, 5);
		gtk_widget_show(album_label);

		album_entry = gtk_entry_new_with_max_length(30);
		gtk_table_attach(GTK_TABLE(table), album_entry, 1, 4, 2, 3, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);
		gtk_widget_show(album_entry);

		year_label = gtk_label_new("Year:");
		gtk_misc_set_alignment(GTK_MISC(year_label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), year_label, 0, 1, 3, 4, GTK_FILL, GTK_FILL, 5, 5);
		gtk_widget_show(year_label);

		year_entry = gtk_entry_new_with_max_length(4);
		gtk_widget_set_usize(year_entry, 40, -1);
		gtk_table_attach(GTK_TABLE(table), year_entry, 1, 2, 3, 4, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);
		gtk_widget_show(year_entry);

		comment_label = gtk_label_new("Comment:");
		gtk_misc_set_alignment(GTK_MISC(comment_label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), comment_label, 2, 3, 3, 4, GTK_FILL, GTK_FILL, 5, 5);
		gtk_widget_show(comment_label);

		comment_entry = gtk_entry_new_with_max_length(30);
		gtk_table_attach(GTK_TABLE(table), comment_entry, 3, 4, 3, 4, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);
		gtk_widget_show(comment_entry);

		genre_label = gtk_label_new("Genre:");
		gtk_misc_set_alignment(GTK_MISC(genre_label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), genre_label, 0, 1, 4, 5, GTK_FILL, GTK_FILL, 5, 5);
		gtk_widget_show(genre_label);

		genre_combo = gtk_combo_new();
		gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(genre_combo)->entry), FALSE);
		if (!genre_list)
		{
			for (i = 0; i < sizeof (id3_genres) / sizeof (*id3_genres); i++)
				genre_list = g_list_append(genre_list, id3_genres[i]);
			genre_list = g_list_sort(genre_list, genre_comp_func);
		}
		gtk_combo_set_popdown_strings(GTK_COMBO(genre_combo), genre_list);

		gtk_table_attach(GTK_TABLE(table), genre_combo, 1, 4, 4, 5, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);
		gtk_widget_show(genre_combo);

		gtk_widget_show(frame);
		gtk_widget_show(table);

		bbox = gtk_hbutton_box_new();
		gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_END);
		gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 5);
		gtk_box_pack_start(GTK_BOX(vbox), bbox, FALSE, FALSE, 0);

		save = gtk_button_new_with_label("Save");
		gtk_signal_connect(GTK_OBJECT(save), "clicked", GTK_SIGNAL_FUNC(save_cb), NULL);
		GTK_WIDGET_SET_FLAGS(save, GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(bbox), save, TRUE, TRUE, 0);
		gtk_widget_show(save);
		gtk_widget_grab_default(save);

		remove_id3 = gtk_button_new_with_label("Remove ID3");
		gtk_signal_connect(GTK_OBJECT(remove_id3), "clicked", GTK_SIGNAL_FUNC(remove_id3_cb), NULL);
		GTK_WIDGET_SET_FLAGS(remove_id3, GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(bbox), remove_id3, TRUE, TRUE, 0);
		gtk_widget_show(remove_id3);

		cancel = gtk_button_new_with_label("Cancel");
		gtk_signal_connect_object(GTK_OBJECT(cancel), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(window));
		GTK_WIDGET_SET_FLAGS(cancel, GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(bbox), cancel, TRUE, TRUE, 0);
		gtk_widget_show(cancel);

		gtk_widget_show(bbox);
		gtk_widget_show(vbox);
		gtk_widget_show(window);
	}

	if (current_filename)
		g_free(current_filename);
	current_filename = g_strdup(filename);

	name = strrchr(filename, '/');
	if (name)
		name++;
	else
		name = filename;

	title = g_strdup_printf("File Info - %s", name);
	gtk_window_set_title(GTK_WINDOW(window), title);
	g_free(title);

	title = g_strdup(name);
	if (tmp = strrchr(title, '.'))
		*tmp = '\0';
	gtk_entry_set_text(GTK_ENTRY(title_entry), title);
	g_free(title);

	gtk_entry_set_text(GTK_ENTRY(artist_entry), "");
	gtk_entry_set_text(GTK_ENTRY(album_entry), "");
	gtk_entry_set_text(GTK_ENTRY(year_entry), "");
	gtk_entry_set_text(GTK_ENTRY(comment_entry), "");
	gtk_list_select_item(GTK_LIST(GTK_COMBO(genre_combo)->list), g_list_index(genre_list, id3_genres[0]));

	if ((fd = open(current_filename, O_RDONLY)) != -1)
	{
		lseek(fd, -sizeof (tag), SEEK_END);
		if (read(fd, &tag, sizeof (tag)) == sizeof (tag))
		{
			if (!strncmp(tag.tag, "TAG", 3))
			{
				set_entry_tag(GTK_ENTRY(title_entry), tag.title, 30);
				set_entry_tag(GTK_ENTRY(artist_entry), tag.artist, 30);
				set_entry_tag(GTK_ENTRY(album_entry), tag.album, 30);
				set_entry_tag(GTK_ENTRY(year_entry), tag.year, 4);
				set_entry_tag(GTK_ENTRY(comment_entry), tag.comment, 30);
				gtk_list_select_item(GTK_LIST(GTK_COMBO(genre_combo)->list), g_list_index(genre_list, id3_genres[tag.genre]));
			}
		}
		close(fd);
	}
}
