/*	xmms - esound output plugin
 *	Copyright (C) 1999	Galex Yen
 *	
 *	this program is free software
 *	
 *	Description:
 *		This program is an output plugin for xmms v0.9 or greater.
 *		The program uses the esound daemon to output audio in order
 *		to allow more than one program to play audio on the same
 *		device at the same time.
 *
 *		Contains code Copyright (C) 1998-1999 Mikael Alm, Olle Hallnas,
 *		Thomas Nillson and 4Front Technologies
 */
 
#include "esdout.h"
#include "config.h"

#define min(x,y) ((x)<(y)?(x):(y))

static gboolean going = FALSE;
static gboolean paused = FALSE;
static gboolean remove_prebuffer = FALSE;
static gint	flush;
static gint	bps;
static AFormat	format;
static gint	channels;
static gint	frequency;
static gint	wr_index=0,rd_index=0;
static gint	output_time_offset;
static void	*buffer;
static gint	buffer_size,prebuffer_size;
static gboolean	prebuffer;
static gint	fd,latency;
static gint	written=0,output_bytes=0;
static gint	length;
static gchar	*hostname = NULL;

static pthread_t buffer_thread;
static pthread_mutex_t flush_mutex = PTHREAD_MUTEX_INITIALIZER, buffer_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t free_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t flush_cond = PTHREAD_COND_INITIALIZER, buffer_cond = PTHREAD_COND_INITIALIZER;
static pthread_cond_t free_cond = PTHREAD_COND_INITIALIZER;

gint esdout_get_written_time(void)
{
	if(!going)
		return 0;
	return (gint)(((gfloat)written*1000)/(gfloat)(bps));
}

gint esdout_get_output_time(void)
{
	gint time;
	
	time = output_time_offset+(gint)((gfloat)((output_bytes - latency)*1000.0)/(gfloat)bps);
	if(time < 0)
		time = 0;
	
	return time;
}

gint esdout_used(void)
{
/*	printf("wr_index : %d   rd_index : %d\n", wr_index, rd_index);*/
	if(wr_index>=rd_index)
		return wr_index-rd_index;
	return buffer_size-(rd_index-wr_index);
}

gint esdout_playing(void)
{
	if (!esdout_used()&&length<=0)	
		return FALSE;
		
	return TRUE;	
}

gint esdout_free(void)
{
	if(remove_prebuffer&&prebuffer)
	{
		prebuffer=FALSE;
		remove_prebuffer=FALSE;
	}
	if(prebuffer)
		remove_prebuffer=TRUE;
	
	if(rd_index>wr_index)
		return rd_index-wr_index;
	return buffer_size-(wr_index-rd_index);
}

void esdout_write(void *ptr,gint length) 
{
	gint cnt,off=0;
	
	if(esdout_free() < length)
	{
		pthread_mutex_lock(&free_mutex);
		while(esdout_free() < length)
			pthread_cond_wait(&free_cond,&free_mutex);
		pthread_mutex_unlock(&free_mutex);
	}
	
	written+=length;
	remove_prebuffer=FALSE;
	while(length>0)
	{
		cnt=min(length,buffer_size-wr_index);
		memcpy(buffer+wr_index,ptr+off,cnt);
		wr_index=(wr_index+cnt)%buffer_size;
		length-=cnt;
		off=cnt;
	}
	pthread_mutex_lock(&buffer_mutex);
	pthread_cond_signal(&buffer_cond);
	pthread_mutex_unlock(&buffer_mutex);

}

void esdout_close(void)
{
	wr_index = 0;
	rd_index = 0;
	going = 0;
	pthread_mutex_lock(&buffer_mutex);
	pthread_cond_signal(&buffer_cond);
	pthread_mutex_unlock(&buffer_mutex);

	pthread_join(buffer_thread,NULL);
	if(hostname)
		g_free(hostname);
	hostname = NULL;
}

void esdout_flush(gint time)
{
	pthread_mutex_lock(&flush_mutex);
	flush=time;
	while(flush!=-1) 
		pthread_cond_wait(&flush_cond,&flush_mutex);
	pthread_mutex_unlock(&flush_mutex);
	
}

void esdout_pause(short p)
{
	paused = p;
	pthread_mutex_lock(&buffer_mutex);
	pthread_cond_signal(&buffer_cond);
	pthread_mutex_unlock(&buffer_mutex);

}

int get_latency(void)
{
	int fd, amount = 0;
#ifndef HAVE_ESD_GET_LATENCY
	esd_server_info_t *info;
#endif
	
	fd = esd_open_sound(hostname);
	
	if(fd == -1)
		return 0;
	
#ifdef HAVE_ESD_GET_LATENCY
	amount = esd_get_latency(fd);
#else
	info = esd_get_server_info(fd);
	if(info)
	{
		if(info->format & ESD_STEREO)
		{
			if(info->format & ESD_BITS16)
				amount = (44100 * (ESD_BUF_SIZE + 64)) / info->rate;
			else
				amount = (44100 * (ESD_BUF_SIZE + 128)) / info->rate;
		}
		else
		{
			if(info->format & ESD_BITS16)
				amount = (2 * 44100 * (ESD_BUF_SIZE + 128)) / info->rate;
			else
				amount = (2 * 44100 * (ESD_BUF_SIZE + 256)) / info->rate;
		}
		free(info);
	}
	amount += ESD_BUF_SIZE * 2;
#endif
	esd_close(fd);
	return amount;
}


void *esdout_loop(void *arg)
{
	gint cnt;
	char *tempbuffer;
	
	
	/* there is this pesky problem with esd... when I start playing a song 
	a loud popping sound often occurs.  This buffer of 0's seems to help.
	*/
	tempbuffer=g_malloc0(ESD_BUF_SIZE*4);
	memset(tempbuffer,0,ESD_BUF_SIZE*4);
	write(fd,tempbuffer,ESD_BUF_SIZE*4);
	g_free(tempbuffer);
	
	
	while(going)
	{
		if(esdout_used() > prebuffer_size) 
			prebuffer = FALSE;
		if (esdout_used() > 0 && !paused && !prebuffer)
		{
			length=min(4096,esdout_used());

			while (length>0)
			{
				cnt=min(length,buffer_size-rd_index);

				output_bytes+=write(fd,buffer+rd_index,cnt);					

				rd_index=(rd_index+cnt)%buffer_size;
				length-=cnt;
			}
			if(!esdout_used())
				prebuffer = TRUE;
			pthread_mutex_lock(&free_mutex);
			pthread_cond_signal(&free_cond);
			pthread_mutex_unlock(&free_mutex);
		}
		else
		{
			pthread_mutex_lock(&buffer_mutex);
			pthread_cond_wait(&buffer_cond,&buffer_mutex);
			pthread_mutex_unlock(&buffer_mutex);
		}
		if(flush != -1)
		{
			output_time_offset = flush;
			written = (flush/10)*(bps/100);
			rd_index = wr_index = output_bytes = 0;
			pthread_mutex_lock(&flush_mutex);
			flush = -1;
			pthread_cond_signal(&flush_cond);
			pthread_mutex_unlock(&flush_mutex);
			prebuffer = TRUE;
		}
	}
	
	pthread_mutex_lock(&free_mutex);
	pthread_cond_signal(&free_cond);
	pthread_mutex_unlock(&free_mutex);
	
	esd_close(fd);
	fd = 0;
	
	g_free(buffer);
	pthread_exit(NULL);
}

int esdout_open(AFormat fmt,gint rate,gint nch)
{
	esd_format_t esdfmt;

	bps=rate*nch;
	if ((fmt==FMT_U16_LE) || (fmt==FMT_U16_BE) || (fmt==FMT_U16_NE) || (fmt==FMT_S16_LE) || 
		(fmt==FMT_S16_BE) || (fmt==FMT_S16_NE)) {
		bps*=2;
	}

	format=fmt;	
	frequency=rate;
	channels=nch;
		
	buffer_size=(esd_cfg.buffer_size*bps)/1000;
	if(buffer_size<8192)
		buffer_size=8192;
	
	prebuffer_size=(buffer_size*esd_cfg.prebuffer)/100;
	if(buffer_size-prebuffer_size<4096)
		prebuffer_size=buffer_size-4096;
	
	buffer=g_malloc0(buffer_size);
	
	going=1;
	flush=-1;
	prebuffer=1;
	wr_index=rd_index=output_time_offset=written=output_bytes=0;
	paused=FALSE;
	remove_prebuffer=FALSE;
	length=0;
	
	if(hostname)
		g_free(hostname);
	if(esd_cfg.use_remote)
		hostname = g_strdup_printf("%s:%d",esd_cfg.server,esd_cfg.port);
	else
		hostname = NULL;
	
	esdfmt = ESD_PLAY;
	if (channels-1)
		esdfmt |= ESD_STEREO;
	else
		esdfmt |= ESD_MONO;
	
	if ((format==FMT_U8) || (format==FMT_S8))
		esdfmt |= ESD_BITS8;
	else
		esdfmt |= ESD_BITS16;
	
	latency = ((get_latency() * frequency)/44100) * channels;
	if(format != FMT_U8 && format != FMT_S8)
		latency *= 2;
	
	fd = esd_play_stream(esdfmt, frequency, hostname, "xmms - plugin");
	if(fd==-1)
	{
		g_free(buffer);
		if(hostname)
			g_free(hostname);
		hostname = NULL;
		return 0;
	}
	
	pthread_create(&buffer_thread,NULL,esdout_loop,NULL);
	return 1;
}

